"""
Utilities - Aggregation Methods

Meteorological data provided by Meteostat (https://dev.meteostat.net)
under the terms of the Creative Commons Attribution-NonCommercial
4.0 International Public License.

The code is licensed under the MIT license.
"""

import numpy as np
import pandas as pd


def weighted_average(step: pd.DataFrame) -> pd.DataFrame:
    """
    Calculate weighted average from grouped data
    """

    data = np.ma.masked_array(step, np.isnan(step))
    data = np.ma.average(data, axis=0, weights=data[:, -1])
    data = data.filled(np.NaN)

    return pd.DataFrame(data=[data], columns=step.columns)


def degree_mean(data: pd.Series) -> float:
    """
    Return the mean of a list of degrees
    """

    rads = np.deg2rad(data)
    sums = np.arctan2(np.sum(np.sin(rads)), np.sum(np.cos(rads)))
    return (np.rad2deg(sums) + 360) % 360



"""
Core Class - Data Loader

Meteorological data provided by Meteostat (https://dev.meteostat.net)
under the terms of the Creative Commons Attribution-NonCommercial
4.0 International Public License.

The code is licensed under the MIT license.
"""

from urllib.error import HTTPError
from multiprocessing.pool import ThreadPool
from typing import Callable
import pandas as pd
from meteostat.core.warn import warn


def processing_handler(
    datasets: list,
    load: Callable[[dict], None],
    max_threads: int
) -> None:
    """
    Load multiple datasets simultaneously
    """

    # Single-thread processing
    if max_threads < 2:

        for dataset in datasets:
            load(*dataset)

    # Multi-thread processing
    else:

        pool = ThreadPool(max_threads)
        pool.starmap(load, datasets)

        # Wait for Pool to finish
        pool.close()
        pool.join()


def load_handler(
    endpoint: str,
    path: str,
    columns: list,
    types: dict,
    parse_dates: list,
    coerce_dates: bool = False
) -> pd.DataFrame:
    """
    Load a single CSV file into a DataFrame
    """

    try:

        # Read CSV file from Meteostat endpoint
        df = pd.read_csv(
            endpoint + path,
            compression='gzip',
            names=columns,
            dtype=types,
            parse_dates=parse_dates)

        # Force datetime conversion
        if coerce_dates:
            df.iloc[:, parse_dates] = df.iloc[:, parse_dates].apply(
                pd.to_datetime, errors='coerce')

    except (FileNotFoundError, HTTPError):

        # Create empty DataFrane
        df = pd.DataFrame(columns=[*types])

        # Display warning
        warn(f'Cannot load {path} from {endpoint}')

    # Return DataFrame
    return df



"""
Base Interface Class

Meteorological data provided by Meteostat (https://dev.meteostat.net)
under the terms of the Creative Commons Attribution-NonCommercial
4.0 International Public License.

The code is licensed under the MIT license.
"""


class Base:

    """
    Base class that provides features which are used across the package
    """

    # Import configuration
    from meteostat.core.config import endpoint, cache_dir, max_age, max_threads



"""
Convert Data Units

Meteorological data provided by Meteostat (https://dev.meteostat.net)
under the terms of the Creative Commons Attribution-NonCommercial
4.0 International Public License.

The code is licensed under the MIT license.
"""

from copy import copy


def convert(
    self,
    units: dict
):
    """
    Convert columns to a different unit
    """

    # Create temporal instance
    temp = copy(self)

    # Change data units
    for parameter, unit in units.items():
        if parameter in temp._columns:
            temp._data[parameter] = temp._data[parameter].apply(unit)

    # Return class instance
    return temp



"""
Point Class

Meteorological data provided by Meteostat (https://dev.meteostat.net)
under the terms of the Creative Commons Attribution-NonCommercial
4.0 International Public License.

The code is licensed under the MIT license.
"""

from datetime import datetime
import pandas as pd
from meteostat.interface.stations import Stations


class Point:

    """
    Automatically select weather stations by geographic location
    """

    # The interpolation method (weighted or nearest)
    method: str = 'weighted'

    # Maximum radius for nearby stations
    radius: int = 35000

    # Maximum difference in altitude
    alt_range: int = 350

    # Maximum number of stations
    max_count: int = 4

    # Adapt temperature data based on altitude
    adapt_temp: bool = True

    # Distance Weight
    weight_dist: float = 0.6

    # Altitude Weight
    weight_alt: float = 0.4

    # The latitude
    _lat: float = None

    # The longitude
    _lon: float = None

    # The altitude
    _alt: int = None

    def __init__(
        self,
        lat: float,
        lon: float,
        alt: int = None
    ) -> None:

        self._lat = lat
        self._lon = lon
        self._alt = alt

        if alt is None:
            self.adapt_temp = False

    def get_stations(self, freq: str, start: datetime,
                     end: datetime) -> pd.DataFrame:
        """
        Get list of nearby weather stations
        """

        # Get nearby weather stations
        stations = Stations()
        stations = stations.nearby(self._lat, self._lon, self.radius)

        # Guess altitude if not set
        if self._alt is None:
            self._alt = stations.fetch().head(self.max_count)[
                'elevation'].mean()

        # Captue unfiltered weather stations
        unfiltered = stations.fetch()
        unfiltered = unfiltered[abs(self._alt -
                                    unfiltered['elevation']) <= self.alt_range]

        # Apply inventory filter
        stations = stations.inventory(freq, (start, end))

        # Apply altitude filter
        stations = stations.fetch()
        stations = stations[abs(self._alt -
                                stations['elevation']) <= self.alt_range]

        # Fill up stations
        selected: int = len(stations.index)
        if selected < self.max_count:
            stations = stations.append(
                unfiltered.head(
                    self.max_count - selected))

        # Calculate score values
        stations['score'] = ((1 - (stations['distance'] / self.radius)) * self.weight_dist) + (
            (1 - (abs(self._alt - stations['elevation']) / self.alt_range)) * self.weight_alt)

        # Sort by score (descending)
        stations = stations.sort_values('score', ascending=False)

        return stations.head(self.max_count)

    @property
    def alt(self) -> int:
        """
        Returns the point's altitude
        """

        # Return altitude
        return self._alt



"""
Aggregate Data

Meteorological data provided by Meteostat (https://dev.meteostat.net)
under the terms of the Creative Commons Attribution-NonCommercial
4.0 International Public License.

The code is licensed under the MIT license.
"""

from copy import copy
import pandas as pd
from meteostat.core.warn import warn


def aggregate(
    self,
    freq: str = None,
    spatial: bool = False
):
    """
    Aggregate observations
    """

    if self.count() > 0 and not self._data.isnull().values.all():

        # Create temporal instance
        temp = copy(self)

        # Set default frequency if not set
        if freq is None:
            freq = self._freq

        # Time aggregation
        temp._data = temp._data.groupby(['station', pd.Grouper(
            level='time', freq=freq)]).agg(temp.aggregations)

        # Spatial aggregation
        if spatial:
            temp._data = temp._data.groupby(
                [pd.Grouper(level='time', freq=freq)]).mean()

        # Round
        temp._data = temp._data.round(1)

        # Return class instance
        return temp

    # Show warning & return self
    warn('Skipping aggregation on empty DataFrame')
    return self



"""
Monthly Interface Class

Meteorological data provided by Meteostat (https://dev.meteostat.net)
under the terms of the Creative Commons Attribution-NonCommercial
4.0 International Public License.

The code is licensed under the MIT license.
"""

from datetime import datetime
from typing import Union
import numpy as np
import pandas as pd
from meteostat.core.cache import get_file_path, file_in_cache
from meteostat.core.loader import processing_handler, load_handler
from meteostat.utilities.validations import validate_series
from meteostat.utilities.aggregations import degree_mean, weighted_average
from meteostat.interface.base import Base
from meteostat.interface.point import Point


class Monthly(Base):

    """
    Retrieve monthly weather data for one or multiple weather stations or
    a single geographical point
    """

    # The cache subdirectory
    cache_subdir: str = 'monthly'

    # The list of weather Stations
    _stations: pd.Index = None

    # The start date
    _start: datetime = None

    # The end date
    _end: datetime = None

    # Include model data?
    _model: bool = True

    # The data frame
    _data: pd.DataFrame = pd.DataFrame()

    # Default frequency
    _freq: str = '1MS'

    # Columns
    _columns: list = [
        'year',
        'month',
        'tavg',
        'tmin',
        'tmax',
        'prcp',
        'snow',
        'wdir',
        'wspd',
        'wpgt',
        'pres',
        'tsun'
    ]

    # Data tapes
    _types: dict = {
        'tavg': 'float64',
        'tmin': 'float64',
        'tmax': 'float64',
        'prcp': 'float64',
        'snow': 'float64',
        'wdir': 'float64',
        'wspd': 'float64',
        'wpgt': 'float64',
        'pres': 'float64',
        'tsun': 'float64'
    }

    # Columns for date parsing
    _parse_dates: dict = {
        'time': [0, 1]
    }

    # Default aggregation functions
    aggregations: dict = {
        'tavg': 'mean',
        'tmin': 'min',
        'tmax': 'max',
        'prcp': 'sum',
        'snow': 'mean',
        'wdir': degree_mean,
        'wspd': 'mean',
        'wpgt': 'max',
        'pres': 'mean',
        'tsun': 'sum'
    }

    def _load(
        self,
        station: str
    ) -> None:
        """
        Load file from Meteostat
        """

        # File name
        file = 'monthly/' + ('full' if self._model else 'obs') + \
            '/' + station + '.csv.gz'

        # Get local file path
        path = get_file_path(self.cache_dir, self.cache_subdir, file)

        # Check if file in cache
        if self.max_age > 0 and file_in_cache(path, self.max_age):

            # Read cached data
            df = pd.read_pickle(path)

        else:

            # Get data from Meteostat
            df = load_handler(
                self.endpoint,
                file,
                self._columns,
                self._types,
                self._parse_dates)

            # Validate Series
            df = validate_series(df, station)

            # Save as Pickle
            if self.max_age > 0:
                df.to_pickle(path)

        # Filter time period and append to DataFrame
        if self._start and self._end:

            # Get time index
            time = df.index.get_level_values('time')

            # Filter & append
            self._data = self._data.append(
                df.loc[(time >= self._start) & (time <= self._end)])

        else:

            # Append
            self._data = self._data.append(df)

    def _get_data(self) -> None:
        """
        Get all required data
        """

        if len(self._stations) > 0:

            # List of datasets
            datasets = []

            for station in self._stations:
                datasets.append((
                    str(station),
                ))

            # Data Processing
            processing_handler(datasets, self._load, self.max_threads)

        else:

            # Empty DataFrame
            self._data = pd.DataFrame(columns=[*self._types])

    def _resolve_point(
        self,
        method: str,
        stations: pd.DataFrame,
        alt: int,
        adapt_temp: bool
    ) -> None:
        """
        Project weather station data onto a single point
        """

        if self._stations.size == 0:
            return None

        if method == 'nearest':

            self._data = self._data.groupby(
                pd.Grouper(level='time', freq='1D')).agg('first')

        else:

            # Join score and elevation of involved weather stations
            data = self._data.join(
                stations[['score', 'elevation']], on='station')

            # Adapt temperature-like data based on altitude
            if adapt_temp:
                data.loc[data['tavg'] != np.NaN, 'tavg'] = data['tavg'] + \
                    ((2 / 3) * ((data['elevation'] - alt) / 100))
                data.loc[data['tmin'] != np.NaN, 'tmin'] = data['tmin'] + \
                    ((2 / 3) * ((data['elevation'] - alt) / 100))
                data.loc[data['tmax'] != np.NaN, 'tmax'] = data['tmax'] + \
                    ((2 / 3) * ((data['elevation'] - alt) / 100))

            # Exclude non-mean data & perform aggregation
            excluded = data['wdir']
            excluded = excluded.groupby(
                pd.Grouper(level='time', freq='1D')).agg('first')

            # Aggregate mean data
            data = data.groupby(
                pd.Grouper(level='time', freq='1D')).apply(weighted_average)

            # Drop RangeIndex
            data.index = data.index.droplevel(1)

            # Merge excluded fields
            data['wdir'] = excluded

            # Drop score and elevation
            self._data = data.drop(['score', 'elevation'], axis=1).round(1)

        # Set placeholder station ID
        self._data['station'] = 'XXXXX'
        self._data = self._data.set_index(
            ['station', self._data.index.get_level_values('time')])
        self._stations = pd.Index(['XXXXX'])

    def __init__(
        self,
        loc: Union[pd.DataFrame, Point, list, str],
        start: datetime = None,
        end: datetime = None,
        model: bool = True
    ) -> None:

        # Set list of weather stations
        if isinstance(loc, pd.DataFrame):
            self._stations = loc.index
        elif isinstance(loc, Point):
            stations = loc.get_stations('hourly', start, end)
            self._stations = stations.index
        else:
            if not isinstance(loc, list):
                loc = [loc]

            self._stations = pd.Index(loc)

        # Set start date
        self._start = start

        # Set end date
        self._end = end

        # Set model
        self._model = model

        # Get data for all weather stations
        self._get_data()

        # Interpolate data
        if isinstance(loc, Point):
            self._resolve_point(loc.method, stations, loc.alt, loc.adapt_temp)

        # Clear cache
        if self.max_age > 0:
            self.clear_cache()

    def expected_rows(self) -> int:
        """
        Return the number of rows expected for the defined date range
        """

        return ((self._end.year - self._start.year) * 12 +
                self._end.month - self._start.month) + 1

    # Import methods
    from meteostat.series.normalize import normalize
    from meteostat.series.interpolate import interpolate
    from meteostat.series.aggregate import aggregate
    from meteostat.series.convert import convert
    from meteostat.series.coverage import coverage
    from meteostat.series.count import count
    from meteostat.series.fetch import fetch
    from meteostat.series.stations import stations
    from meteostat.core.cache import clear_cache



"""
Normalize Data

Meteorological data provided by Meteostat (https://dev.meteostat.net)
under the terms of the Creative Commons Attribution-NonCommercial
4.0 International Public License.

The code is licensed under the MIT license.
"""

from copy import copy
from numpy import NaN
import pandas as pd
from meteostat.core.warn import warn


def normalize(self):
    """
    Normalize the DataFrame
    """

    if self.count() == 0:
        warn('Pointless normalization of empty DataFrame')

    # Create temporal instance
    temp = copy(self)

    # Create result DataFrame
    result = pd.DataFrame(columns=temp._columns[1:])

    # Go through list of weather stations
    for station in temp._stations:
        # Create data frame
        df = pd.DataFrame(columns=temp._columns[1:])
        # Add time series
        df['time'] = pd.date_range(temp._start, temp._end, freq=self._freq)
        # Add station ID
        df['station'] = station
        # Add columns
        for column in temp._columns[1:]:
            # Add column to DataFrame
            df[column] = NaN

        result = pd.concat([result, df], axis=0)

    # Set index
    result = result.set_index(['station', 'time'])

    # Merge data
    temp._data = pd.concat([temp._data, result], axis=0).groupby(
        ['station', 'time'], as_index=True).first()

    # None -> NaN
    temp._data = temp._data.fillna(NaN)

    # Return class instance
    return temp



"""
Utilities - Validations

Meteorological data provided by Meteostat (https://dev.meteostat.net)
under the terms of the Creative Commons Attribution-NonCommercial
4.0 International Public License.

The code is licensed under the MIT license.
"""

import pandas as pd


def validate_series(
    df: pd.DataFrame,
    station: str
) -> pd.DataFrame:
    """
    Make sure a series is formatted correctly
    """

    # Add missing column(s)
    if 'time' not in df.columns:
        df['time'] = None

    # Add weather station ID
    df['station'] = station

    # Set index
    df = df.set_index(['station', 'time'])

    # Return DataFrame
    return df



"""
Core Class - Warnings

Meteorological data provided by Meteostat (https://dev.meteostat.net)
under the terms of the Creative Commons Attribution-NonCommercial
4.0 International Public License.

The code is licensed under the MIT license.
"""

import warnings


def _warning_on_one_line(message, category, *args) -> str:  # pylint: disable=unused-argument
    """
    Print warning on a single line
    """
    return '%s: %s\n' % (category.__name__, message)


# Set warning format
warnings.formatwarning = _warning_on_one_line


def warn(message: str) -> None:
    """
    Create a warning
    """
    warnings.warn(message, Warning)



"""
Core Class - Configuration

Meteorological data provided by Meteostat (https://dev.meteostat.net)
under the terms of the Creative Commons Attribution-NonCommercial
4.0 International Public License.

The code is licensed under the MIT license.
"""

import os

# Base URL of the Meteostat bulk data interface
endpoint: str = 'https://bulk.meteostat.net/v2/'

# Location of the cache directory
cache_dir: str = os.path.expanduser(
    '~') + os.sep + '.meteostat' + os.sep + 'cache'

# Maximum age of a cached file in seconds
max_age: int = 24 * 60 * 60

# Maximum number of threads used for downloading files
max_threads: int = 1



"""
█▀▄▀█ █▀▀ ▀█▀ █▀▀ █▀█ █▀ ▀█▀ ▄▀█ ▀█▀
█░▀░█ ██▄ ░█░ ██▄ █▄█ ▄█ ░█░ █▀█ ░█░

A Python library for accessing open weather and climate data

Meteorological data provided by Meteostat (https://dev.meteostat.net)
under the terms of the Creative Commons Attribution-NonCommercial
4.0 International Public License.

The code is licensed under the MIT license.
"""

__appname__ = 'meteostat'
__version__ = '1.3.0'

from .interface.stations import Stations
from .interface.point import Point
from .interface.hourly import Hourly
from .interface.daily import Daily
from .interface.monthly import Monthly
from .interface.normals import Normals
from .utilities import units



"""
Normals Interface Class

Meteorological data provided by Meteostat (https://dev.meteostat.net)
under the terms of the Creative Commons Attribution-NonCommercial
4.0 International Public License.

The code is licensed under the MIT license.
"""

from datetime import datetime
from typing import Union
import pandas as pd
from meteostat.utilities.aggregations import degree_mean
from meteostat.interface.base import Base
from meteostat.interface.point import Point
from meteostat.interface.monthly import Monthly


class Normals(Base):

    """
    Retrieve climate normals for one or multiple weather stations or
    a single geographical point
    """

    # The list of weather Stations
    _stations: pd.Index = None

    # The data frame
    _data: pd.DataFrame = pd.DataFrame()

    # Columns
    _columns = Monthly._columns

    # Default aggregation functions
    aggregations: dict = {
        'tavg': 'mean',
        'tmin': 'mean',
        'tmax': 'mean',
        'prcp': 'mean',
        'snow': 'mean',
        'wdir': degree_mean,
        'wspd': 'mean',
        'wpgt': 'mean',
        'pres': 'mean',
        'tsun': 'mean'
    }

    # Coverage
    _coverage: dict = {}

    def __init__(
        self,
        loc: Union[pd.DataFrame, Point, list, str],
        start: int = 1981,
        end: int = 2020,
        model: bool = True
    ) -> None:

        # Convert start & end to datetime
        start = datetime(start, 1, 1)
        end = datetime(end, 12, 31)

        # Get monthly data
        raw = Monthly(loc, start, end, model)

        # Get list of weather stations
        self._stations = raw.stations

        # Get DataFrame
        self._data = raw._data

        # Get coverage
        self._coverage['global'] = raw.coverage()
        for parameter in self._columns[2:]:
            self._coverage[parameter] = raw.coverage(parameter)

        # Aggregate
        self._data = self._data.groupby(['station', self._data.index.get_level_values(
            'time').month]).agg(self.aggregations)

        # Rename time column
        self._data.index.rename(['station', 'month'], inplace=True)

    def coverage(
        self,
        parameter: str = None
    ) -> float:
        """
        Return data coverage (overall or by parameter)
        """

        if parameter is None:
            return self._coverage['global']

        return self._coverage[parameter]

    # Import methods
    from meteostat.series.convert import convert
    from meteostat.series.count import count
    from meteostat.series.fetch import fetch



"""
Get Weather Stations

Meteorological data provided by Meteostat (https://dev.meteostat.net)
under the terms of the Creative Commons Attribution-NonCommercial
4.0 International Public License.

The code is licensed under the MIT license.
"""

from copy import copy
import pandas as pd


@property
def stations(self) -> pd.Index:
    """
    Fetch Weather Stations
    """

    # Return index of weather stations
    return copy(self._stations)
